/////////////////////////////////////////////////////////////////////////////
//
//	MG35TOOL
//		This program will split MG35 firmware into it's separate component
//		parts, and join them back into a firmware image.
//
//	Author:		Jeff Curless
//
//	Information for this application was taken from:
//		http://mediagate.pbwiki.com/MG35Firmware
//
/////////////////////////////////////////////////////////////////////////////

// mg35tool.cpp : Defines the entry point for the console application.
//

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "mg35tool.h"

#if !defined(WIN32)
#define _stricmp strcasecmp
#endif

struct CmdInfo Command;
unsigned char *Buffer;

/////////////////////////////////////////////////////////////////////////////
//
//	WriteFile(  char *name, unsigned char *Buffer, DWORD Length )
//
//	This routine writes out a file from the buffer passed, using the length
//	parameter for the sized of the file.
//
//	Returns:
//		true if successful.
//
/////////////////////////////////////////////////////////////////////////////

bool
WriteFile( char *name, unsigned char *Buffer, DWORD Length )
{
	FILE   *fp;
	bool	bResult = false;

	fp = fopen( name, "wb" );
	if( fp ){
		if( fwrite( Buffer, 1, Length, fp ) == Length ){
			bResult = true;
		}
		fclose( fp );
	}

	return bResult;
}

/////////////////////////////////////////////////////////////////////////////
//
//	ReadFile( char *name, BYTE *Buffer, DWORD Length )
//
//	This routine reads Length bytes from the file passed in, into buffer
//
//	Returns:
//		true if successful.
//
/////////////////////////////////////////////////////////////////////////////

bool
ReadFile( char *name, BYTE *Buffer, DWORD Length )
{
	FILE   *fp;
	bool	bResult = false;

	fp = fopen( name, "rb" );
	if( fp ){
		if( fread( Buffer, 1, Length, fp ) == Length ){
			bResult = true;
		}
		fclose( fp );
	}

	return bResult;
}

/////////////////////////////////////////////////////////////////////////////
//
//	FileLength( FILE *fp  )
//
//	This routine calculates the size of a file, when the file is already 
//	open.
//
//	Returns:
//		The size of the file in bytes, or zero if error.
//
//	NOTE:
//		This routine *WILL* reset the current file pointer location back to
//		the start of the file.
//
/////////////////////////////////////////////////////////////////////////////

DWORD
FileLength( FILE *fp  )
{
	DWORD	size = 0;

	if( fp ){
		fseek( fp, 0, SEEK_END );
		size = ftell( fp );
		fseek( fp, 0, SEEK_SET );
	}

	return size;
}


/////////////////////////////////////////////////////////////////////////////
//
//	GetFileLength
//
//	This routine determines the length of a file in bytes
//	
//	Returns:
//		true if successful, if so size contains the actual size of the file
//		in bytes.
//
/////////////////////////////////////////////////////////////////////////////

bool
GetFileLength( char *name, DWORD *size )
{
	FILE	*fp;
	*size = 0;

	fp = fopen( name, "rb" );
	if( fp ){
		fseek( fp, 0, SEEK_END );
		*size = ftell( fp );
		fseek( fp, 0, SEEK_SET );
		fclose( fp );
		return true;
	}
	else{
		printf( "Could not open file %s\n", name );
	}

	return false;
}

/////////////////////////////////////////////////////////////////////////////
//
//	CalculateChecksum( DWORD_LE *Memory, DWORD Length )
//
//	This routine calculates the checksum of section of memory, the checksum
//	is the sum of all DWORDS within that memory region, so the length passed
//  in *MUST* be a count of the DWORDS in the memory region.
//
//	Returns
//		The checksum.
/////////////////////////////////////////////////////////////////////////////

DWORD
CalculateChecksum( DWORD_LE *Memory, DWORD Length )
{
	DWORD i;
	DWORD Sum = 0;

	for( i = 0 ; i < Length ; i++ ){
		Sum += Memory[i];
	}

	return Sum;
}

/////////////////////////////////////////////////////////////////////////////
//
//	IsChecksumValid( DWORD_LE *Memory, DWORD Length, DWORD Checksum )
//
//	This routine determines if the passed in checksum is equal to the
//	checksum generated when we calculate it for the memory region passed.
//
//	Returns:
//		true if checksums match.
//
/////////////////////////////////////////////////////////////////////////////

bool
IsChecksumValid( DWORD_LE *Memory, DWORD Length, DWORD Checksum )
{
	DWORD Sum;

	//
	// We calculate the checksum based on DWORD sizes, so.. convert length to
	// number of DWORDS.
	//
	Sum = CalculateChecksum( Memory, Length/sizeof(DWORD) );

	if( Sum == Checksum ){
		return( true );
	}
	else{
		return( false );
	}
}

/////////////////////////////////////////////////////////////////////////////
//
//	ProcessBootSection( CInfo *pCmd, struct BOOTHeader *Boot )
//
//	This routine processes the boot section of the firmware image, if one
//	exists.
//	
//	Returns:
//		true if successful, if so the requested boot.img file is written out.
//
/////////////////////////////////////////////////////////////////////////////
bool
ProcessBootSection( CInfo *pCmd, struct BOOTHeader *Boot )
{
	bool	bResult = false;

	printf( "[BOOT]\n");
	printf( "  Checksum         = 0x%08X\n",(DWORD)Boot->Checksum);
	printf( "  Offset in Flash  = 0x%08X\n",(DWORD)Boot->OffsetToBootLoaderInFlash);
	printf( "  Length           = 0x%08X\n",(DWORD)Boot->Length);

	//
	// Validate the boot loader checksum
	//
	if( IsChecksumValid( (DWORD_LE *)&Boot->DataStart, Boot->Length, Boot->Checksum ) ){
		//
		// Write out the boot loader
		//
		printf( "  Checksum matches, generating boot.img output\n\n");
		if( WriteFile( pCmd->BootImage, &Boot->DataStart, Boot->Length ) ){
			bResult = true;
		}
		else{
			printf("Could not write %s, aborting process\n", pCmd->BootImage);
		}
	}
	else{
		printf("Checksum of boot image does not match checksum in header, aborting\n");
	}

	return bResult;
}

bool
ProcessBootSection25( CInfo *pCmd, struct BOOTHeader25 *Boot )
{
	bool	bResult = false;

	printf( "[BOOT]\n");
	printf( "  Checksum         = 0x%08X\n",(DWORD)Boot->Checksum);
	printf( "  Offset in Flash  = 0x%08X\n",(DWORD)Boot->OffsetToBootLoaderInFlash);
	printf( "  Length           = 0x%08X\n",(DWORD)Boot->Length);

	//
	// Validate the boot loader checksum
	//
	if( IsChecksumValid( (DWORD_LE *)&Boot->DataStart, Boot->Length, Boot->Checksum ) ){
		//
		// Write out the boot loader
		//
		printf( "  Checksum matches, generating boot.img output\n\n");
		if( WriteFile( pCmd->BootImage, &Boot->DataStart, Boot->Length ) ){
			bResult = true;
		}
		else{
			printf("Could not write %s, aborting process\n", pCmd->BootImage);
		}
	}
	else{
		printf("Checksum of boot image does not match checksum in header, aborting\n");
	}

	return bResult;
}


/////////////////////////////////////////////////////////////////////////////
//
//	ProcessRofsSection( CInfo *pCmd, struct ROFSHeader *Rofs )
//
//	This routine processes the section of the firmware file that contains
//	the ROFS header, and generates the images for the cramfs and linux kernel
//
/////////////////////////////////////////////////////////////////////////////

bool
ProcessRofsSection( CInfo *pCmd, struct ROFSHeader *Rofs )
{
	bool  bResult = false;
	DWORD Count;

	printf( "[ROFS]\n" );
	printf( "  CheckSum         = 0x%08X\n",(DWORD)Rofs->Checksum);
	printf( "  Offset in Flash  = 0x%08X\n",(DWORD)Rofs->OffsetInFlash);
	printf( "  Length           = 0x%08X\n",(DWORD)Rofs->Length);
	printf( "  Offset to Cramfs = 0x%08X\n",(DWORD)Rofs->OffsetToCramfs);

	if( IsChecksumValid( (DWORD_LE *)&Rofs->OffsetToCramfs,Rofs->Length,Rofs->Checksum ) ){

		Count = (DWORD)Rofs->OffsetToCramfs - (DWORD)Rofs->OffsetInFlash - 4;
		if( WriteFile( pCmd->KernelImage, &Rofs->DataStart, Count ) ){
			if( WriteFile( pCmd->CramfsImage, &Rofs->DataStart+Count, (DWORD)Rofs->Length - Count - 4 ) ){
				bResult = true;
			}
			else{
				printf( "Could not write file %s\n", pCmd->CramfsImage );
			}
		}
		else{
			printf( "Could not write to file l%s\n", pCmd->KernelImage );
		}
	}
	else{
		printf( "Checksum of ROFS section does not match checksum in header, aborting\n" );
	}

	return bResult;
}

bool
ProcessRofsSection25( CInfo *pCmd, struct ROFSHeader25 *Rofs )
{
	bool  bResult = false;
	DWORD Count;

	printf( "[ROFS]\n" );
	printf( "  CheckSum         = 0x%08X\n",(DWORD)Rofs->Checksum);
	printf( "  Offset in Flash  = 0x%08X\n",(DWORD)Rofs->OffsetInFlash);
	printf( "  Length           = 0x%08X\n",(DWORD)Rofs->Length);
	printf( "  Offset to Cramfs = 0x%08X\n",(DWORD)Rofs->OffsetToCramfs);

	if( IsChecksumValid( (DWORD_LE *)&Rofs->OffsetToCramfs,Rofs->Length,Rofs->Checksum ) ){

		Count = (DWORD)Rofs->OffsetToCramfs - (DWORD)Rofs->OffsetInFlash - 4;
		if( WriteFile( pCmd->KernelImage, &Rofs->DataStart, Count ) ){
			if( WriteFile( pCmd->CramfsImage, &Rofs->DataStart+Count, (DWORD)Rofs->Length - Count - 4 ) ){
				bResult = true;
			}
			else{
				printf( "Could not write file %s\n", pCmd->CramfsImage );
			}
		}
		else{
			printf( "Could not write to file l%s\n", pCmd->KernelImage );
		}
	}
	else{
		printf( "Checksum of ROFS section does not match checksum in header, aborting\n" );
	}

	return bResult;
}


/////////////////////////////////////////////////////////////////////////////
//
//	SplitMemoryImage
// 
//	Have in memory image of .upgrade file need to split it out into it's 
//  component parts:
//
//  <taken from http://mediagate.pbwiki.com/MG35Firmware>
//  [UPGIXM02] (ASCII)
//  [VERS]     (Dword) - Firmware version (ie: 03-04-01-00 = v1.4.3)
//  
//  [BOOT]     (ASCII) - Optional section
//  [CKSMb]    (Dword) - Checksum of bootloader
//  [OFFSb]    (Dword) - Offset to bootloader in flash (0x00000)
//  [BOOTLEN]  (Dword) - Length of bootloader
//  [BOOTLDR]  (Data)  - * Start of the bootloader data
//  ...
//  
//  [ROFS]     (ASCII)
//  [CKSMr]    (Dword) - Checksum of ROFS
//  [OFFSr]    (Dword) - Offset to ROFS in flash (0x40000)
//  [ROFSLEN]  (Dword) - Length of ROFS ([OFFSc]+[LINUX]+[CRAMFS])
//  [OFFSc]    (Dword) - * Offset to Cramfs.img in flash
//  [LINXGZ]   (Data)  - Start of the linux.gz data
//  ...
//  [CRAMFS]   (Data)  - Start of the Cramfs.img data
//  ...
//
//  (* = Section checksum calculation starts here)
//  (Note: All data is in little-endian byte order)
//	
////////////////////////////////////////////////////////////////////////////

bool
SplitMemoryImage( CInfo *pCmd, unsigned char *Buffer )
{
	DWORD				  Version;
	DWORD				  Length;
	bool				  bResult = false;

	if ( pCmd->bMG25 )
	{
		struct UpgradeHeader25 *Header;
		struct ROFSHeader25	 *Rofs;
		struct BOOTHeader25	 *Boot;

		Header = (struct UpgradeHeader25 *)Buffer;

		//
		// Check header tag to see if the file is really a .upgrade file
		//
		if( strncmp( Header->Tag, pCmd->FirmwareTag, strlen(pCmd->FirmwareTag)) ){
			printf( "Firmware does not have a valid header, aborting\n" );
			printf( "Header is: %8s\n",Header->Tag );
		}
		else{
			//
			// Provide the version number
			//  
			Version = Header->Version;
			printf( "Firmware version is %d.%d.%d\n",(Version>>16)&0xFF,(Version>>8)&0xFF,(Version&0xFF));
			printf( "MG25 Upgrade file size is %d\n", Header->FileSize);

			//
			// What is the next section, BOOT or ROFS?
			//
			if( !strncmp( Header->NextSection, MG35_BOOT_STRING, strlen(MG35_BOOT_STRING)) ){
				//
				// Process the boot section
				//
				Boot = (struct BOOTHeader25 *)&Header->NextSection;
				if( ProcessBootSection25( pCmd, Boot ) ){
					Length = Boot->Length;

					//
					// Process the rofs section if there is one
					//
					Rofs = (struct ROFSHeader25 *)&Boot->DataStart;
					Rofs = (struct ROFSHeader25 *)(((char *)Rofs) + Length);
					if( !strncmp( Rofs->RofsSection, MG35_ROFS_STRING, strlen(MG35_ROFS_STRING)) ){
						if( ProcessRofsSection25( pCmd, Rofs ) ){
							bResult = true;
						}
					}
				}
			}
			else if( !strncmp( Header->NextSection, MG35_ROFS_STRING, strlen(MG35_ROFS_STRING)) ){
				Rofs = (struct ROFSHeader25 *)&Header->NextSection;
				if( ProcessRofsSection25( pCmd, Rofs ) ){
					bResult = true;
				}
			}
			else{
				printf( "Located unknown section [%-4s] in image, aborting\n",Header->NextSection);
			}
		}
	}
	else
	{
		struct UpgradeHeader *Header;
		struct ROFSHeader	 *Rofs;
		struct BOOTHeader	 *Boot;

		Header = (struct UpgradeHeader *)Buffer;

		//
		// Check header tag to see if the file is really a .upgrade file
		//
		if( strncmp( Header->Tag, pCmd->FirmwareTag, strlen(pCmd->FirmwareTag)) ){
			printf( "Firmware does not have a valid header, aborting\n" );
			printf( "Header is: %8s\n",Header->Tag );
		}
		else{
			//
			// Provide the version number
			//  
			Version = Header->Version;
			printf( "Firmware version is %d.%d.%d\n",(Version>>16)&0xFF,(Version>>8)&0xFF,(Version&0xFF));

			//
			// What is the next section, BOOT or ROFS?
			//
			if( !strncmp( Header->NextSection, MG35_BOOT_STRING, strlen(MG35_BOOT_STRING)) ){
				//
				// Process the boot section
				//
				Boot = (struct BOOTHeader *)&Header->NextSection;
				if( ProcessBootSection( pCmd, Boot ) ){
					Length = Boot->Length;

					//
					// Process the rofs section if there is one
					//
					Rofs = (struct ROFSHeader *)&Boot->DataStart;
					Rofs = (struct ROFSHeader *)(((char *)Rofs) + Length);
					if( !strncmp( Rofs->RofsSection, MG35_ROFS_STRING, strlen(MG35_ROFS_STRING)) ){
						if( ProcessRofsSection( pCmd, Rofs ) ){
							bResult = true;
						}
					}
				}
			}
			else if( !strncmp( Header->NextSection, MG35_ROFS_STRING, strlen(MG35_ROFS_STRING)) ){
				Rofs = (struct ROFSHeader *)&Header->NextSection;
				if( ProcessRofsSection( pCmd, Rofs ) ){
					bResult = true;
				}
			}
			else{
				printf( "Located unknown section [%-4s] in image, aborting\n",Header->NextSection);
			}
		}
	}

	return bResult;
}

/////////////////////////////////////////////////////////////////////////////
//
//	SplitImageFile( CInfo *pCmd )
//
//	This routine starts the process of splitting a firmware image into it's
//	component parts.
//
//	Returns:
//		true if successful, and if successful generated files for each section
//		within the firmware image.
//
/////////////////////////////////////////////////////////////////////////////

bool
SplitImageFile( CInfo *pCmd )
{
	FILE *fp;
	bool	bResult = false;
	int		size;

	fp = fopen( pCmd->UpdateFile, "rb" );
	if( fp ){
		size = FileLength( fp );

		//
		// Obtain a buffer and read the file
		//
		Buffer = (unsigned char *)malloc( size );
		if( Buffer ){
			if( size == fread( Buffer, 1, size, fp ) ){
				bResult = SplitMemoryImage( pCmd, Buffer );
			}
			else{
				printf( "Error reading file\n" );
			}
			free( Buffer );
		}
		else{
			printf( "Could not allocated sufficient resources to read in file\n" );
		}
		fclose( fp );
	}
	else{
		printf( "Could not open file %s\n", pCmd->UpdateFile );
	}

	return bResult;
}

/////////////////////////////////////////////////////////////////////////////
//
//	FillInBootSection
//
//	Generate the boot section of the firmware image
//
//  [BOOT]     (ASCII) - Optional section
//  [CKSMb]    (Dword) - Checksum of bootloader
//  [OFFSb]    (Dword) - Offset to bootloader in flash (0x00000)
//  [BOOTLEN]  (Dword) - Length of bootloader
//  [BOOTLDR]  (Data)  - * Start of the bootloader data
//
//	Returns:
//		true if successful, and the BOOTHeader section is filled in, along
//		with a copy of the boot image given.
//
/////////////////////////////////////////////////////////////////////////////

bool
FillInBootSection( CInfo *pCmd, struct BOOTHeader *pBoot )
{
	bool  bResult = false;
	DWORD LengthInDWORDS;

	memcpy( pBoot->BootSection, MG35_BOOT_STRING, strlen( MG35_BOOT_STRING ) );
	pBoot->OffsetToBootLoaderInFlash = pCmd->BootOffsetInFlash;
	pBoot->Length = pCmd->BootSize;

	if( ReadFile( pCmd->BootImage, &pBoot->DataStart, pCmd->BootSize ) ){
		LengthInDWORDS = pBoot->Length/sizeof(DWORD);
		pBoot->Checksum = CalculateChecksum( (DWORD_LE *)&pBoot->DataStart, LengthInDWORDS );
		printf( "[BOOT]\n");
		printf( "  CheckSum         = 0x%08X\n",(DWORD)pBoot->Checksum);
		printf( "  Offset in Flash  = 0x%08X\n",(DWORD)pBoot->OffsetToBootLoaderInFlash);
		printf( "  Length           = 0x%08X\n",(DWORD)pBoot->Length);
		bResult = true;
	}
	else{
		printf("Could not read in boot image file %s\n",pCmd->BootImage);
	}

	return bResult;
}

bool
FillInBootSection25( CInfo *pCmd, struct BOOTHeader25 *pBoot )
{
	bool  bResult = false;
	DWORD LengthInDWORDS;

	memcpy( pBoot->BootSection, MG35_BOOT_STRING, strlen( MG35_BOOT_STRING ) );
	pBoot->OffsetToBootLoaderInFlash = pCmd->BootOffsetInFlash;
	pBoot->Length = pCmd->BootSize;

	if( ReadFile( pCmd->BootImage, &pBoot->DataStart, pCmd->BootSize ) ){
		LengthInDWORDS = pBoot->Length/sizeof(DWORD);
		pBoot->Checksum = CalculateChecksum( (DWORD_LE *)&pBoot->DataStart, LengthInDWORDS );
		printf( "[BOOT]\n");
		printf( "  CheckSum         = 0x%08X\n",(DWORD)pBoot->Checksum);
		printf( "  Offset in Flash  = 0x%08X\n",(DWORD)pBoot->OffsetToBootLoaderInFlash);
		printf( "  Length           = 0x%08X\n",(DWORD)pBoot->Length);
		bResult = true;
	}
	else{
		printf("Could not read in boot image file %s\n",pCmd->BootImage);
	}

	return bResult;
}


/////////////////////////////////////////////////////////////////////////////
//
//	FillInRofsSection( CInfo *pCmd, struct ROFSHeader *Rofs)
//
//	Generate the ROFS section of the firmware image.
//
//	[ROFS]     (ASCII)
//  [CKSMr]    (Dword) - Checksum of ROFS
//  [OFFSr]    (Dword) - Offset to ROFS in flash (0x40000)
//  [ROFSLEN]  (Dword) - Length of ROFS ([OFFSc]+[LINUX]+[CRAMFS])
//  [OFFSc]    (Dword) - * Offset to Cramfs.img in flash
//  [LINXGZ]   (Data)  - Start of the linux.gz data
//  ...
//  [CRAMFS]   (Data)  - Start of the Cramfs.img data
//
//	Returns:
//		true if successful, and the ROFSHeader is fully populated with the
//		linux kernel and cramfs info and data.
//
/////////////////////////////////////////////////////////////////////////////

bool
FillInRofsSection( CInfo *pCmd, struct ROFSHeader *pRofs )
{
	bool	bResult = false;
	DWORD	LengthInDWORDS;
	
	memcpy( pRofs->RofsSection, MG35_ROFS_STRING, sizeof( MG35_ROFS_STRING ) );
	pRofs->OffsetInFlash  = pCmd->ROFSOffsetInFlash;
	pRofs->Length         = pCmd->KernelSize + pCmd->CramfsSize + sizeof(DWORD);
	pRofs->OffsetToCramfs = pRofs->OffsetInFlash + pCmd->KernelSize + sizeof(DWORD);
	if( ReadFile( pCmd->KernelImage, &pRofs->DataStart, pCmd->KernelSize ) ){
		if( ReadFile( pCmd->CramfsImage, &pRofs->DataStart + pCmd->KernelSize, pCmd->CramfsSize ) ){
			LengthInDWORDS = (pCmd->KernelSize + pCmd->CramfsSize + sizeof(DWORD))/sizeof(DWORD);
			pRofs->Checksum = CalculateChecksum( &pRofs->OffsetToCramfs, LengthInDWORDS );

			printf( "[ROFS]\n");
			printf( "  CheckSum         = 0x%08X\n",(DWORD)pRofs->Checksum);
			printf( "  Offset in Flash  = 0x%08X\n",(DWORD)pRofs->OffsetInFlash);
			printf( "  Length           = 0x%08X\n",(DWORD)pRofs->Length);
			printf( "  Offset to Cramfs = 0x%08X\n",(DWORD)pRofs->OffsetToCramfs);

			bResult = true;
		}
	}

	return bResult;
}

bool
FillInRofsSection25( CInfo *pCmd, struct ROFSHeader25 *pRofs )
{
	bool	bResult = false;
	DWORD	LengthInDWORDS;
	
	memcpy( pRofs->RofsSection, MG35_ROFS_STRING, sizeof( MG35_ROFS_STRING ) );
	pRofs->OffsetInFlash  = pCmd->ROFSOffsetInFlash;
	pRofs->Length         = pCmd->KernelSize + pCmd->CramfsSize + sizeof(DWORD);
	pRofs->OffsetToCramfs = pRofs->OffsetInFlash + pCmd->KernelSize + sizeof(DWORD);
	if( ReadFile( pCmd->KernelImage, &pRofs->DataStart, pCmd->KernelSize ) ){
		if( ReadFile( pCmd->CramfsImage, &pRofs->DataStart + pCmd->KernelSize, pCmd->CramfsSize ) ){
			LengthInDWORDS = (pCmd->KernelSize + pCmd->CramfsSize + sizeof(DWORD))/sizeof(DWORD);
			pRofs->Checksum = CalculateChecksum( &pRofs->OffsetToCramfs, LengthInDWORDS );

			printf( "[ROFS]\n");
			printf( "  CheckSum         = 0x%08X\n",(DWORD)pRofs->Checksum);
			printf( "  Offset in Flash  = 0x%08X\n",(DWORD)pRofs->OffsetInFlash);
			printf( "  Length           = 0x%08X\n",(DWORD)pRofs->Length);
			printf( "  Offset to Cramfs = 0x%08X\n",(DWORD)pRofs->OffsetToCramfs);

			bResult = true;
		}
	}

	return bResult;
}

/////////////////////////////////////////////////////////////////////////////
//
//	JoinMemoryImage( CInfo *pCmd, BYTE *Buffer, DWORD *Length )
//
//  This routine does the setup work, then calls down to functions to fill
//  in the rest of the firmware image, depending on what the user wanted the
//  image to contain.
//
//  Returns:
//		true if successful, and Length contains the actual amount (in bytes)
//		used within the Buffer.
/////////////////////////////////////////////////////////////////////////////

bool
JoinMemoryImage( CInfo *pCmd, BYTE *Buffer, DWORD *Length )
{
	DWORD				  Used = 0;
	bool				  bResult = false;

	if ( pCmd->bMG25 )
	{
		struct UpgradeHeader25 *pHeader;
		struct BOOTHeader25    *pBoot;
		struct ROFSHeader25    *pRofs;

		//
		// Build up header
		//
		pHeader = (struct UpgradeHeader25 *)Buffer;
		memcpy( pHeader->Tag, pCmd->FirmwareTag, strlen( pCmd->FirmwareTag ) );
		Used = (DWORD)strlen(pCmd->FirmwareTag);

		pHeader->Version = pCmd->Version;
		pHeader->FileSize = 0;
		Used += 8;
		
		//
		// If the user specified a -b, generate an image with a boot loader, if not
		// generate one without a boot loader
		// 
		if( pCmd->bBoot ){
			printf( "*** WARNING ***\nThis image contains a boot loader, if there is a problem flashing your\n");
			printf( "system, you may not be able to recover it.\n");
			pBoot = (struct BOOTHeader25 *)&pHeader->NextSection;
			if( FillInBootSection25( pCmd, pBoot ) ){
				Used += sizeof( struct BOOTHeader25 ) - 1;
				Used += pCmd->BootSize;
				if( FillInRofsSection25( pCmd, (struct ROFSHeader25 *)((BYTE *)pHeader + Used) ) ){
					Used += pCmd->KernelSize;
					Used += pCmd->CramfsSize;
					Used += sizeof( struct ROFSHeader25 ) - 1;
					pHeader->FileSize = Used;
					bResult = true;
				}
				else{
					printf( "Error importing rofs section, aborting\n");
				}
			}
			else{
				printf( "Error importing boot section, aborting\n");
			}
		}
		else{
			pRofs = (struct ROFSHeader25 *)&pHeader->NextSection;
			if( FillInRofsSection25( pCmd, pRofs ) ){
				Used += pCmd->KernelSize;
				Used += pCmd->CramfsSize;
				Used += sizeof( struct ROFSHeader25 ) - 1;
				pHeader->FileSize = Used;
				bResult = true;
			}
		}
	}
	else
	{
		struct UpgradeHeader *pHeader;
		struct BOOTHeader    *pBoot;
		struct ROFSHeader    *pRofs;

		//
		// Build up header
		//
		pHeader = (struct UpgradeHeader *)Buffer;
		memcpy( pHeader->Tag, pCmd->FirmwareTag, strlen( pCmd->FirmwareTag ) );
		Used = (DWORD)strlen(pCmd->FirmwareTag);

		pHeader->Version = pCmd->Version;
		Used += 4;
		
		//
		// If the user specified a -b, generate an image with a boot loader, if not
		// generate one without a boot loader
		// 
		if( pCmd->bBoot ){
			printf( "*** WARNING ***\nThis image contains a boot loader, if there is a problem flashing your\n");
			printf( "system, you may not be able to recover it.\n");
			pBoot = (struct BOOTHeader *)&pHeader->NextSection;
			if( FillInBootSection( pCmd, pBoot ) ){
				Used += sizeof( struct BOOTHeader ) - 1;
				Used += pCmd->BootSize;
				if( FillInRofsSection( pCmd, (struct ROFSHeader *)((BYTE *)pHeader + Used) ) ){
					Used += pCmd->KernelSize;
					Used += pCmd->CramfsSize;
					Used += sizeof( struct ROFSHeader ) - 1;
					bResult = true;
				}
				else{
					printf( "Error importing rofs section, aborting\n");
				}
			}
			else{
				printf( "Error importing boot section, aborting\n");
			}
		}
		else{
			pRofs = (struct ROFSHeader *)&pHeader->NextSection;
			if( FillInRofsSection( pCmd, pRofs ) ){
				Used += pCmd->KernelSize;
				Used += pCmd->CramfsSize;
				Used += sizeof( struct ROFSHeader ) - 1;
				bResult = true;
			}
		}
	}

	if( bResult ){
		*Length = Used;
	}

	return bResult;
}

/////////////////////////////////////////////////////////////////////////////
//
//	JoinImageFile( CInfo *pCmd )
//
//  Need to build in memory image of the firmware, then write it out.
//
//  <taken from http://mediagate.pbwiki.com/MG35Firmware>
//
//  [UPGIXM02] (ASCII)
//  [VERS]     (Dword) - Firmware version (ie: 03-04-01-00 = v1.4.3)
//  
//  [BOOT]     (ASCII) - Optional section
//  [CKSMb]    (Dword) - Checksum of bootloader
//  [OFFSb]    (Dword) - Offset to bootloader in flash (0x00000)
//  [BOOTLEN]  (Dword) - Length of bootloader
//  [BOOTLDR]  (Data)  - * Start of the bootloader data
//  ...
//  
//  [ROFS]     (ASCII)
//  [CKSMr]    (Dword) - Checksum of ROFS
//  [OFFSr]    (Dword) - Offset to ROFS in flash (0x40000)
//  [ROFSLEN]  (Dword) - Length of ROFS ([OFFSc]+[LINUX]+[CRAMFS])
//  [OFFSc]    (Dword) - * Offset to Cramfs.img in flash
//  [LINXGZ]   (Data)  - Start of the linux.gz data
//  ...
//  [CRAMFS]   (Data)  - Start of the Cramfs.img data
//  ...
//
//  (* = Section checksum calculation starts here)
//  (Note: All data is in little-endian byte order)
//
//	Returns:
//		true if successful.
//
/////////////////////////////////////////////////////////////////////////////

bool
JoinImageFile( CInfo *pCmd )
{
	DWORD	FileSize;
	unsigned char *Buffer = NULL;
	bool	bResult = false;

	//
	// Get file sizes!
	//
	if( pCmd->bBoot ){
		if( !GetFileLength( pCmd->BootImage, &pCmd->BootSize ) ){
			return bResult;
		}
	}
	else{
		pCmd->BootSize = 0;
	}

	if( GetFileLength( pCmd->KernelImage, &pCmd->KernelSize ) ){
		if( GetFileLength( pCmd->CramfsImage, &pCmd->CramfsSize ) ){
			//
			// Note this calculation wil be larger than what is needed, but not by much
			//
			FileSize  = pCmd->BootSize + pCmd->KernelSize + pCmd->CramfsSize;
			FileSize += sizeof( struct UpgradeHeader );
			FileSize += sizeof( struct ROFSHeader );
			if( pCmd->bBoot ){
				FileSize += sizeof( struct BOOTHeader );
			}

			//
			// Create the needed buffer, and start filling it in.
			//
			Buffer = (unsigned char *)malloc(FileSize);
			if( Buffer ){
				if( JoinMemoryImage( pCmd, Buffer, &FileSize ) ){
					if( WriteFile( pCmd->UpdateFile, Buffer, FileSize ) ){
						bResult = true;
						printf( "join operation successful.\n");
					}
					else{
						printf( "Could not write .update file\n" );
					}
				}
				else{
					printf( "Could not join files into .update file\n");
				}
				free( Buffer );
			}
			else{
				printf( "Could not allocate enough memory to hold firmware image, aborting\n");
			}
		}
	}

	return bResult;
}

/////////////////////////////////////////////////////////////////////////////
//
//	ConvertVersionToDWORD( char *Version )
//
//	Convert the passed in string, into a DWORD version number that can
//  be used later.  All incorrect parseing will generate a version number
//  of 1.4.4
//
//	Returns:
//		The version number in a DWORD, or 0x00010404 if not successful.
//
/////////////////////////////////////////////////////////////////////////////

DWORD
ConvertVersionToDWORD( char *Version )
{
	DWORD	Major;
	DWORD	Minor;
	DWORD	Edit;
	DWORD	Result = 0;
	bool	bHaveResult = false;

	if( sscanf( Version, "%d.%d.%d", &Major,&Minor,&Edit) == 3 ){
		bHaveResult = true;
	}
	else if( sscanf( Version, "v%d.%d.%d", &Major,&Minor,&Edit ) == 3 ){
		bHaveResult = true;
	}
	else if( sscanf( Version, "V%d.%d.%d", &Major,&Minor,&Edit ) == 3 ){
		bHaveResult = true;
	}

	if( bHaveResult ){
		Result  = (Major&0xFF) << 16;
		Result |= (Minor&0xFF) << 8;
		Result |= (Edit&0xFF);
	}
	else{
		printf( "Illegal syntax for version field, setting to 1.4.4\n");
		Result = 0x00010404;
	}

	return Result;
}

/////////////////////////////////////////////////////////////////////////////
//
//	ProcessArguments( CInfo *pCmd, int argc, char *argv[] )
//
//	Process command line arguments, setting up the pCmd structure.  This
//  structure will be used to help process the commands later.
//
//  Note:  The order of the arguments are important, as long as the switch
//         values all have the needed argument immediately after them.
//
//	Returns:
//		true if the parse was successful, and does not contain conflicting 
//		requests.
//
/////////////////////////////////////////////////////////////////////////////

bool
ProcessArguments( CInfo *pCmd, int argc, char *argv[] )
{
	int		i;
	bool	bValidArgument = false;
	bool	bSyntaxError = false;

	i = 1;
	while( i < argc ){
		if( !_stricmp( argv[i], "split" ) ){
			pCmd->bSplit = true;
		}
		else if( !_stricmp( argv[i], "join" ) ){
			pCmd->bJoin = true;
		}
		else if( !_stricmp( argv[i], "-b" ) ){
			pCmd->bBoot = true;
			if( ((i+1) < argc) && (argv[i+1][0] != '-') ){
				pCmd->BootImage = argv[i+1];
			}
			else{
				bSyntaxError = true;
			}
			i++;
		}
		else if( !_stricmp( argv[i], "-k" ) ){
			if( ((i+1) < argc) && (argv[i+1][0] != '-') ){
				pCmd->KernelImage = argv[i+1];
			}
			else{
				bSyntaxError = true;
			}
			i++;
		}
		else if( !_stricmp( argv[i], "-c" ) ){
			if( ((i+1) < argc) && (argv[i+1][0] != '-') ){
				pCmd->CramfsImage = argv[i+1];
			}
			else{
				bSyntaxError = true;
			}
			i++;
		}
		else if( !_stricmp( argv[i], "-v" ) ){
			if( ((i+1) < argc) && (argv[i+1][0] != '-') ){
				pCmd->Version = ConvertVersionToDWORD(argv[i+1]);
			}
			else{
				bSyntaxError = true;
			}
			i++;
		}
		else if( !_stricmp( argv[i], "-t" ) ){
			if( ((i+1) < argc) && (argv[i+1][0] != '-') ){
				if( !_stricmp( argv[i+1], "MG35" ) ){
					pCmd->FirmwareTag = MG35_TAG_STRING;
					pCmd->BootOffsetInFlash = MG35_BOOT_OFFSET;
					pCmd->ROFSOffsetInFlash = MG35_ROFS_OFFSET;
				}
				else if( !_stricmp( argv[i+1], "MG350HD" ) ){
					pCmd->FirmwareTag		= MG350_TAG_STRING;
					pCmd->BootOffsetInFlash = MG350_BOOT_OFFSET;
					pCmd->ROFSOffsetInFlash = MG350_ROFS_OFFSET;
				}
				else if( !_stricmp( argv[i+1], "MG25" ) ){
					pCmd->FirmwareTag		= MG25_TAG_STRING;
					pCmd->BootOffsetInFlash = MG25_BOOT_OFFSET;
					pCmd->ROFSOffsetInFlash = MG25_ROFS_OFFSET;
					pCmd->bMG25 = true;
				}
				else{
					printf( "illegal type specified: %s\n",argv[i+1]);
					bSyntaxError = true;
				}
			}
			else{
				printf( "Requires type parameter\n");
				bSyntaxError = true;
			}
			i++;
		}
		else if( argv[i][0] == '-' ){
			printf( "Illegal switch passed, %s\n",argv[i]);
			bSyntaxError = true;
		}
		else if( (i + 1) == argc ){
			// better be the last argument, and ... the upgrade file
			pCmd->UpdateFile = argv[i];
		}
		else{
			printf( "Illegal argument passed, %s\n", argv[i] );
			bSyntaxError = true;
		}
		i++;
	}

	//
	// Check for a single command
	//
	if( !bSyntaxError ){
		if( pCmd->bJoin ^ pCmd->bSplit ){
			if( pCmd->UpdateFile != NULL ){
				bValidArgument = true;
			}
			else{
				printf( "No firmware upgrade filename given.\n");
			}
		}
		else{
			printf( "You must specify either join, or split\n");
		}
	}

	return bValidArgument;
}

/////////////////////////////////////////////////////////////////////////////
//
//	Usage
//
//	Display proper usage of this tool
//
/////////////////////////////////////////////////////////////////////////////

void
Usage()
{
	printf( "Usage: mg35tool <split | join> [options] <upgrade file>\n\n" );
	printf( "Options are:\n");
	printf( "   -b boot file, default boot.img\n");
	printf( "   -k kernel file, default linux.gz\n");
	printf( "   -c cramfs file, default cramfs.img\n");
	printf( "   -v version of firmware\n");
	printf( "   -t type, one of {MG35, MG350HD, MG25}, MG35 is default.\n");
	printf( "\nExample:\n\n");
	printf( "mg35tool split -k clinux.krn -c oldcramfs.img mg35.1.4.4.update\n\n");
	printf( "mg35tool join -k linux.gz -c mycramfs.img newfirmware.update\n\n");
	printf( "\nNote: If the -b option is not specified on join, no boot image will be added\n");
	printf( "to the firmware image.\n");
}

/////////////////////////////////////////////////////////////////////////////
//
//	main
//
/////////////////////////////////////////////////////////////////////////////

int main(int argc, char * argv[])
{
	if( argc < 2 ){
		Usage();
	}
	else if( ProcessArguments( &Command, argc, argv ) ){
		if( Command.bSplit ){
			SplitImageFile( &Command );
		}
		if( Command.bJoin ){
			JoinImageFile( &Command );
		}
	}

	return 0;
}
